﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Madart.OperationalUI.DataAccess.Models;
using Madart.OperationalUI.DataAccess.Repositories.Interface;
using NAisParser;

namespace Madart.OperationalUI.DataAccess.Repositories
{
    public partial interface IVesselRepository
    {
        Vessel GetVessel(int mmsi);
        int? GetVesselTypeGroupId(int mmsi);
        int? GetVesselTypeId(int mmsi);
		bool UpdateVessel(NAisParser.MessageType5 Update);
		string GetDescription(int mmsiNo);

	}
	public class VesselRepository : AbstractRepository<Vessel>, IVesselRepository
	{
		public Vessel GetVessel(int mmsi)
		{
			long? mmsiLong = mmsi;
			Vessel current = (from vessel in Context.Vessel
							  where vessel.MmsiNo == mmsiLong
							  select vessel).FirstOrDefault();

			return current;
		}

		public int? GetVesselTypeGroupId(int mmsi)
		{
			long? vesselTypeGroupId = (from vessel in Context.Vessel
									   join vesseltype in Context.VesselType on vessel.VesselTypeId equals vesseltype.VesselTypeId
									   where vessel.MmsiNo == mmsi
									   select vesseltype.VesselTypeGroupId).FirstOrDefault();
			return (int?)vesselTypeGroupId;
		}

		public int? GetVesselTypeId(int mmsi)
		{
			long? vesselTypeId = (from vessel in Context.Vessel
								  where vessel.MmsiNo == mmsi
								  select vessel.VesselTypeId).FirstOrDefault();
			return (int?)vesselTypeId;
		}
		public string GetDescription(int mmsi)
		{
			var description = (from vessel in Context.Vessel
							   where vessel.MmsiNo == mmsi
							   select vessel.Description).FirstOrDefault();
			return description;
		}


		/// <summary>
		/// Handles a messageType5. If the message maps to an existing vessel in the database, the vessel is updated. If the vessel does not exist, an new is created with data from the update.
		/// </summary>
		/// <param name="Update">The update-message</param>
		/// <returns>True if the message is handled, false if something goes wrong.</returns>
		public bool UpdateVessel(MessageType5 Update)
		{
			try
			{
				if(Update.ImoNumber == 0)
				{
					var mmsiVessel = Context.Vessel.FirstOrDefault(v => v.MmsiNo == Update.Mmsi);
					if (mmsiVessel == null)
					{
						Vessel newVessel = VesselFromMessage5(Update);
						Context.Vessel.Add(newVessel);
						Context.SaveChanges();
						Context.Entry(newVessel).State = Microsoft.EntityFrameworkCore.EntityState.Detached;
						return true;
					}
					UpdateDBVessel(Update, mmsiVessel);
					Context.SaveChanges();
					return true;

				}

				var imoVessel = Context.Vessel.FirstOrDefault(v => v.ImoNo == Update.ImoNumber);

				if (imoVessel == null)
				{
					Vessel newVessel = VesselFromMessage5(Update);
					Context.Vessel.Add(newVessel);
					Context.SaveChanges();
					Context.Entry(newVessel).State = Microsoft.EntityFrameworkCore.EntityState.Detached;
					return true;

				}
				
				imoVessel.MmsiNo = Update.Mmsi;
				UpdateDBVessel(Update, imoVessel);
				Context.SaveChanges();
				return true;

			}
			catch (Exception e)
			{
				Debug.WriteLine(e.ToString());
				Debug.WriteLine(e.StackTrace);
				return false;
			}
		}

		private static void UpdateDBVessel(MessageType5 Update, Vessel imoVessel)
		{
			imoVessel.VesselName = Update.VesselName;
			imoVessel.CallSign = Update.CallSign;
			var updateDescription = Update.ShipType.ToString();
			if (!string.IsNullOrWhiteSpace(updateDescription))
			{
				if (!string.IsNullOrWhiteSpace(imoVessel.Description) && !imoVessel.Description.Equals(updateDescription)) imoVessel.Description = updateDescription;

			}
		}


		/// <summary>
		/// Creates a new Vessel based on information from a MessageType5 message
		/// </summary>
		/// <param name="update">MessageType5 to map information from</param>
		/// <returns>new Vessel with data from the messageType5</returns>
		private static Vessel VesselFromMessage5(MessageType5 update)
		{
			var newVessel = new Vessel();
			newVessel.MmsiNo = update.Mmsi;
			if (update.ImoNumber == 0)
			{
				newVessel.ImoNo = null;
			}
			else
			{
				newVessel.ImoNo = update.ImoNumber;

			}
			newVessel.VesselName = update.VesselName;
			newVessel.CallSign = update.CallSign;
			newVessel.Description = update.ShipType.ToString();
			return newVessel;
		}

	}
}
