﻿using System;
using System.Collections.Generic;
using Madart.OperationalUI.MadAPI;
using NpgsqlTypes;
using NUnit.Framework;
using Madart.OperationalUI.MadAPI.Models;
using System.Threading.Tasks;
using NAisParser;
using Madart.OperationalUI.Common.Extensions.DataAccess;
using Madart.OperationalUI.DataAccess.Models;
using System.Diagnostics;

namespace Madart.OperationalUI.Tests.ControllersAndHandlers
{
	class MadAPIClientTests
	{
		
		Uri getRoutesUrl = new Uri("http://172.19.10.185:5000/api/get_routes");
		Uri getDeviationsUrl = new Uri("http://172.19.10.185:5000/api/get_deviations");

		private MessageType123 GenerateMessageType123(double longitude, double latitude, int sog, double cog)
		{
			return new MessageType123(new byte(), new byte(), 00000, 0, 0.0, sog, false, longitude, latitude, cog, 0, 0, 0, false, 0);
		}


		[Test]
		public async Task Post_empty_ToGetRoutes()
		{
			string emptyJsonString = "{\"trajectory\":[],\"sog\":[],\"cog\":0}";
			string expectedResult = "{\"result\": -1}";

			string result = await MadApiClient.Post(getRoutesUrl, emptyJsonString);
			Assert.AreEqual(result, expectedResult);
		}

		[Test]
		public async Task Post_empty_ToGetDeviations()
		{
			string emptyJsonString = "{\"lat\": 0, \"lon\": 0, \"sog\": 0, \"cog\": 0, \"normal_route_id\": 0}";

			Assert.AreEqual(await MadApiClient.Post(getDeviationsUrl, emptyJsonString), null);
		}

		[Test]
		public void convertVesselStatesToTH_test()
		{
			List<VesselState> vesselStates = new List<VesselState>();
			TrajectoryHistory trajectoryHistory = new TrajectoryHistory();

			NpgsqlPoint position_a = new NpgsqlPoint(80, 60);
			NpgsqlPoint position_b = new NpgsqlPoint(81, 61);
			NpgsqlPoint position_c = new NpgsqlPoint(82, 62);

			MessageType123 message_a = GenerateMessageType123(position_a.X, position_a.Y, 50, 0.0);
			MessageType123 message_b = GenerateMessageType123(position_b.X, position_b.Y, 100, 0.0);
			MessageType123 message_c = GenerateMessageType123(position_c.X, position_c.Y, 150, 3.0); ;

			VesselState vesselState_a = new VesselState().MapFrom(message_a);
			VesselState vesselState_b = new VesselState().MapFrom(message_b);
			VesselState vesselState_c = new VesselState().MapFrom(message_c);

			vesselStates.Add(vesselState_a);
			vesselStates.Add(vesselState_b);
			vesselStates.Add(vesselState_c);

			float? latestCos = (float)vesselStates[vesselStates.Count - 1].CourseOverGround;
			List<float?> listOfSog = new List<float?>();
			List<List<float>> listOfPos = new List<List<float>>();
			for (int i = 0; i < vesselStates.Count; i++)
			{
				List<float> coord = new List<float>
				{
					(float)vesselStates[i].Position.X,
					(float)vesselStates[i].Position.Y
				};
				listOfPos.Add(coord);
			}


			for (int i = 0; i < vesselStates.Count; i++)
			{
				listOfSog.Add((float)(vesselStates[i].SpeedOverGround));

			}

			trajectoryHistory.trajectory = listOfPos;
			trajectoryHistory.sog = listOfSog;
			trajectoryHistory.cog = latestCos;

			List<float> pos_a = new List<float> { 80, 60 };
			List<float> pos_b = new List<float> { 81, 61 };
			List<float> pos_c = new List<float> { 82, 62 };

			List<List<float>> expectedTrajectory = new List<List<float>> { pos_a, pos_b, pos_c };
			List<float> expectedSog = new List<float> { 50, 100, 150 };
			float expectedCog = 3;
			// string expectedJsonString = "{\"trajectory\":[[80, 60], [81, 61], [82, 62]],\"sog\":[5, 10, 15],\"cog\":3}";

			Assert.AreEqual(trajectoryHistory.trajectory, expectedTrajectory);
			Assert.AreEqual(expectedSog, trajectoryHistory.sog);
			Assert.AreEqual(expectedCog, trajectoryHistory.cog);

		}

		[Test]
		public async Task getNormalRoutes_TH_withOnePoint()
		{
			VesselState vesselState = new VesselState();
			List<float> pos_a = new List<float> { 60, 80 };
			MessageType123 message = new MessageType123(new byte(), new byte(), 00000, 0, 0.0, 5, true, (double)pos_a[0], (double)pos_a[1], 0.0, 0, 0, 0, false, 0);
			VesselStateExtensions.MapFrom(vesselState, message);
			List<VesselState> vesselStates = new List<VesselState>
			{
				vesselState
			};

			Assert.Null(await MadApiClient.GetNormalRoutesAsync(vesselStates));
		}

		[Test]
		public void getNormalRoutes_TH_withMultiplePoints (){

			//{"trajectory":[[5.30299,58.9066849],[5.30385828,58.91093],[5.30477,58.9152756],[5.305675,58.9195557]],"sog":[142.0,141.0,141.0,142.0],"cog":5.0}
			List<List<float>> trajectory_a = new List<List<float>> {
				new List<float> { 15.855052f , 68.3127f },
				new List<float> { 15.8462868f, 68.30915f},
				new List<float> { 15.8418283f, 68.30746f},
				new List<float> { 15.8367786f, 68.30558f},
				new List<float> { 15.8317432f, 68.30374f},
				new List<float> { 15.8263168f, 68.30183f }
			};
			List<float?> sog_a = new List<float?> { 83.0f, 86.0f, 85.0f, 86.0f, 80.0f, 87.0f };
			float cog_a = 225.89999389648438f;
			List<VesselState> vesselStates = new List<VesselState>();
			for (int i = 0; i < trajectory_a.Count; i++)
			{
				MessageType123 message = GenerateMessageType123(trajectory_a[i][0], trajectory_a[i][1], (int)sog_a[i], (double)cog_a);
				VesselState vesselState = new VesselState().MapFrom(message);
				vesselStates.Add(vesselState);
			}

			Assert.NotNull(MadApiClient.GetNormalRoutesAsync(vesselStates));

		}

		[Test]
		public async Task LowDeviationTest()
		{
			NpgsqlPoint position_a = new NpgsqlPoint(12.559811666666667, 67.00487166666667);
			MessageType123 message_a = GenerateMessageType123(position_a.X, position_a.Y, 154, 217.9);
			VesselState vesselState_a = new VesselState().MapFrom(message_a);

			var deviation = await MadApiClient.GetDeviationAsync(vesselState_a, 31190);
			Assert.IsTrue(deviation.rel_dist_max < 0.5);
			
		}
		[Test]
		public async Task HighDeviationTest()
		{
			NpgsqlPoint position_a = new NpgsqlPoint(12.359811666666667, 68.10487166666667);
			MessageType123 message_a = GenerateMessageType123(position_a.X, position_a.Y, 154, 217.9);
			VesselState vesselState_a = new VesselState().MapFrom(message_a);

			var deviation = await MadApiClient.GetDeviationAsync(vesselState_a, 31190);
			Assert.IsTrue(deviation.rel_dist_max >= 3);
		}

		[Test]
		public async Task InDangerOfCoast()
		{
			NpgsqlPoint position_a = new NpgsqlPoint(8.054727, 63.235038);
			MessageType123 message_a = GenerateMessageType123(position_a.X, position_a.Y, 154, 154);
			VesselState vesselState_a = new VesselState().MapFrom(message_a);

			var danger = await MadApiClient.CheckForDangerOfRunningAground(vesselState_a);
			Assert.Greater(danger.c_in_sector??0, 0);


		}

		[Test]
		public async Task InDangerOfSkerry()
		{
			NpgsqlPoint position_a = new NpgsqlPoint(10.459595, 59.068106);
			MessageType123 message_a = GenerateMessageType123(position_a.X, position_a.Y, 154, 180);
			VesselState vesselState_a = new VesselState().MapFrom(message_a);

			var danger = await MadApiClient.CheckForDangerOfRunningAground(vesselState_a);
			Assert.Greater(danger.s_in_sector??0, 0);

		}

		[Test] 
		public async Task InNoImmediateDanger()
		{
			NpgsqlPoint position_a = new NpgsqlPoint(5.868402, 62.452841);
			MessageType123 message_a = GenerateMessageType123(position_a.X, position_a.Y, 154, 1);
			VesselState vesselState_a = new VesselState().MapFrom(message_a);

			var danger = await MadApiClient.CheckForDangerOfRunningAground(vesselState_a);
			Assert.Zero(danger.c_in_sector ?? 0);
		}
	}
}
