﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;
using Madart.OperationalUI.AIS;
using Madart.OperationalUI.Common.Extensions.DataAccess;
using Madart.OperationalUI.DataAccess.Models;
using Madart.OperationalUI.DataAccess.Repositories;
using Madart.OperationalUI.MadAPI;
using Madart.OperationalUI.MadAPI.Models;
using Madart.OperationalUI.WebApp.States;
using Madart.OperationalUI.SignalR.Models;
using Madart.OperationalUI.Hubs;
using Microsoft.AspNetCore.SignalR;
using Microsoft.Extensions.DependencyInjection;


namespace Madart.OperationalUI.WebApp.AIS
{
	public class AisMessageHandler
	{
		protected int processedMessages = 0;
		protected int filteredByDeadband = 0;
		protected int threadOverflowCount = 0;
		protected int nullMessageCount = 0;
		protected object nullSyncRoot = new object();
		protected object counterSyncRoot = new object();
		protected object dbSyncRoot = new object();
		protected int max = 0;
		private IServiceProvider ServiceProvider;
		protected int maxTasks = 24;
		protected ISet<Task> Tasks = new HashSet<Task>();

		/// <summary>
		/// The main event-method, constructs a vesselstates, handles it and posts the results to cliend and database. 
		/// </summary>
		/// <param name="args"> An AisListenerEventArg, containing messagenumber, the MessageType123 parsed from a stream and time received</param>
		/// <param name="vesselStateManager">Interface connected to our vesselstateManager</param>
		/// <param name="hubContext">Interface connected to our hubs, allows the method to post to all clients</param>
		/// <returns>To be implemented: an int/enum that relays the status, if it was discarded/handled, etc</returns>
		public Task HandleMessageAsync(AisMessageEventArgs args, IVesselStateCollection vesselStateManager, IHubContext<VesselHub> hubContext, IServiceProvider serviceProvider)
		{
			ServiceProvider = serviceProvider;
			//IVesselStateRepository VesselRepository = ServiceProvider.GetService<IVesselStateRepository>();
			if (args == null || args.Message == null) return Task.FromResult<object>(null);
			max = args.MessageNo;

			//Parse AIS-message
			var vesselState = new VesselState().MapFrom(args.Message);

			if (vesselState == null) // Means that vessel could not be parsed, or that an error occurred during parsing
			{
				lock (nullSyncRoot)
				{
					nullMessageCount++;
					if (nullMessageCount % 100 == 0)
					{
						Debug.WriteLine($"Number of nullStates: {nullMessageCount}");
					}
				}
				return Task.FromResult<object>(null);
			}
			if (!VesselStateValidator.IsPositionValid(vesselState)) return Task.FromResult<object>(null);
			vesselState.TimeReceived = args.TimeStamp;



			#pragma warning restore CS4014 // Because this call is not awaited, execution of the current method continues before the call is completed
			var NewCurrentVesselState = new CurrentVesselState().MapFrom(vesselState);
			var PreviousCurrentVesselState = vesselStateManager.GetCurrentVesselState(vesselState.MmsiNo);

			bool update = TryUpdate(vesselStateManager, NewCurrentVesselState, PreviousCurrentVesselState);

			#pragma warning disable CS4014 // Because this call is not awaited, execution of the current method continues before the call is completed
			var task = Task.Run(async () =>
			{

				lock (counterSyncRoot)
				{

					processedMessages++;
					if (processedMessages % 100 == 0)
					{
						Debug.WriteLine($"Processed {processedMessages} messages of {max} ({Math.Round(((double)processedMessages / max) * 100.0, 2)}%)");
						Debug.WriteLine($"Timestamp of message that is being processed: {vesselState.TimeReceived}");
					}
				}
				if (!update) // Means that update was not accepted. Could be that movement is too small, too big, the position is out of bounds or that the message is too old
				{
					filteredByDeadband++;
					if (filteredByDeadband % 100 == 0)
					{
						Debug.WriteLine($"{filteredByDeadband} messages filtered by deadband or position validation");
					}
					return;
				}
				try
				{
					Deviation deviation = null;
					PostVesselStateToDBAndBroadcast(hubContext, vesselState, NewCurrentVesselState, new VesselStateRepository());
					if (!WorthCheckingForNormalRoute(NewCurrentVesselState)) // Returns if the vessel is of type ferry, Fish og Tug 
					{
						return;
					}
					WarningCollection.ClearVesselWarning(vesselState.MmsiNo); // Clear previous warning
																		   // Checks if vessel has been classified to a route previously
					var routeIdRanks = vesselStateManager.GetNormalRouteIdRankList(vesselState.MmsiNo);
					if (routeIdRanks == null || routeIdRanks.Count == 0 || routeIdRanks[0].Item1 == -1) // Means that the vessel doesn't have a normal route assigned
					{

						routeIdRanks = await Classify(vesselState.MmsiNo, new VesselStateRepository());
					}

					if (routeIdRanks != null && routeIdRanks.Count > 0 && routeIdRanks[0].Item1 != -1)
					{
						vesselState.NormalRouteId = routeIdRanks[0].Item1;
						vesselStateManager.UpdateNormalroute(vesselState.MmsiNo, routeIdRanks);
					
						// Checks if vessel is off course
						deviation = await GetBestDeviation(vesselState, routeIdRanks);
						if (deviation == null) // Means that Could not get deviation
						{
							Debug.WriteLine("Could not get deviation");
						}


						if (deviation.rel_dist_std >= 6.6)                                        // Deviation is large, try to reclassify
						{

							var oldIdRanks = routeIdRanks;
							var newIdRanks =  await Classify(vesselState.MmsiNo, new VesselStateRepository());         // Tries go get a better matching route
							if (newIdRanks != null && oldIdRanks != newIdRanks)                                               // Got a better suited route, try to get new deviation
							{
								deviation = await GetBestDeviation(vesselState, routeIdRanks);
								UpdateVesselDeviation(vesselState, deviation);
								vesselStateManager.UpdateNormalroute(vesselState.MmsiNo, newIdRanks);
								vesselStateManager.UpdateDeviation(vesselState.MmsiNo, deviation);
								vesselState.NormalRouteId = routeIdRanks[0].Item1;
							}
							if (deviation.rel_dist_std >= 6.6 && vesselState.SpeedOverGround >= 2)                                // Deviation is still too large, check if vessel is in danger of running aground
							{
								CheckForDangerOfRunningAground(hubContext, vesselState);
							}

						}

					}
					if (deviation != null)
					{
						UpdateVesselDeviation(vesselState, deviation);
						vesselStateManager.UpdateDeviation(vesselState.MmsiNo, deviation);

					}
				PostVesselStateToDBAndBroadcast(hubContext, vesselState, NewCurrentVesselState, new VesselStateRepository(), false);

				}
				catch (Exception e)
				{
					Debug.WriteLine(e.Message + ": " + e.StackTrace);
				}
				

			});

			lock (Tasks)
			{
				Tasks.Add(task);
			}
			task.ContinueWith(t =>
			{
				lock (Tasks)
				{
					if (Tasks.Contains(task))
					{
						Tasks.Remove(task);
					}
				}
			});
			return Task.CompletedTask;

		#pragma warning restore CS4014 // Because this call is not awaited, execution of the current method continues before the call is completed
		}

		/// <summary>
		/// Checks what type of vessel currentVessel is, and if it is something we want to keep an eye for
		/// </summary>
		/// <param name="currentVessel">The vessel you want to know if you should check for normalroute</param>
		/// <returns>True if the vessel is of type Cargo, Tanker, Auxiliary or undefined</returns>
		public bool WorthCheckingForNormalRoute(CurrentVesselState currentVessel)
		{
			var IgnoredShipTypes = new List<string>(new string[] {"HighSpeedCraft", "Tug", "Fishing", "Towing", "PilotVessel", "SearchAndRescueVessel","Dredging", "DivingOps", "MedicalTransport" });


			if (currentVessel.VesselTypeGroupId != null)
			{
				if (currentVessel.VesselTypeGroupId == 100233 || currentVessel.VesselTypeGroupId == 100234)
				{
					return false;
				}
				if(currentVessel.VesselTypeGroupId == 100232) // if vesselGroupType = passanger
				{
					var i = currentVessel.VesselTypeId;
					if (i == 101287 || i == 101415) // returns false for VesselType ferry and ro-ro passanger
					{
						return false;
					}
					
				}
				

			}
			if (IgnoredShipTypes.Contains(currentVessel.Description))
			{
				return false;
			}

			if (Tasks.Count >= maxTasks)
			{
				threadOverflowCount++;
				processedMessages--;
				if (threadOverflowCount % 100 == 0)
				{
					Debug.WriteLine($"{threadOverflowCount} messages ignored due to thread overflow");
				}
				return false;
			}

			return true;
		}
		/// <summary>
		/// Fetches history and sends it to the APIClient for classifying of voyage
		/// </summary>
		/// <param name="MmsiNo">The MmsiNo to the vessel that needs classifying</param>
		/// <returns name = "normalRoutes">A list containing tuples with id and rank of the normalroutes the vessel was classified to</returns>
		private async Task<List<Tuple<int?,double?>>> Classify(int MmsiNo, IVesselStateRepository VesselStateRepository)
		{
			var history = VesselStateRepository.GetHistory(MmsiNo);

			// Only get vessel states where speed is larger than 0.1 knots
			history = history.TakeWhile(h => h.SpeedOverGround > 0.1).ToList();
			if (history.Count < 7) return null; // Not enough history to make a classification

			var normalRoutes = await MadApiClient.GetNormalRoutesAsync(history); 
			return normalRoutes;
		}

		/// <summary>
		/// Iterates through the routes assigned to the vessel to fetch the best Deviation at the given point
		/// </summary>
		/// <param name="vesselState">The VesselState we are checking for</param>
		/// <param name="routeIdRanks">The routes we are iterating through and fetching deviation for</param>
		/// <returns name="bestDeviation"> The best deviation from the routes, or the first one under 6.6 RelDist, (for optimization we dont do everyone where do not have to)</returns>
		private static async Task<Deviation> GetBestDeviation(VesselState vesselState, List<Tuple<int?, double?>> routeIdRanks)
		{
			try
			{
				Deviation deviation = new Deviation(); //Should keep a tab on what is the best deviation, even if all of them are over 6.6
				Deviation bestDeviation = await MadApiClient.GetDeviationAsync(vesselState, routeIdRanks[0].Item1);
				if (bestDeviation == null)
				{
					bestDeviation = new Deviation();
					bestDeviation.rel_dist_std = 8;
				}
				foreach (var routeIdRank in routeIdRanks) // Breaks on 6,6 could iterate through everyone and pick the best one, but might take a lot of time
				{
					deviation = await MadApiClient.GetDeviationAsync(vesselState, routeIdRank.Item1);
					if (deviation != null)
					{
						if (deviation.rel_dist_std < bestDeviation.rel_dist_std) bestDeviation = deviation;

					}
				}

				return bestDeviation;
			}catch(Exception e)
			{
				Debug.WriteLine(e.ToString());
				return null;
			}
			
		}

		/// <summary>
		/// Checks if a vessel that is deviating too much is in danger of running aground
		/// </summary>
		/// <param name="hubContext">The hubcontext so the method can post a warning can be sent to all clients</param>
		/// <param name="vesselState">The VesselState that is to be evaluated</param>
		/// <returns></returns>
		private async static Task CheckForDangerOfRunningAground(IHubContext<VesselHub> hubContext, VesselState vesselState)
		{
			var DangerOfRunningAgroundData = await MadApiClient.CheckForDangerOfRunningAground(vesselState);
			if (DangerOfRunningAgroundData != null && (DangerOfRunningAgroundData.c_in_sector != 0 || DangerOfRunningAgroundData.s_in_sector != 0))
			{
				var newWarning = new Warning(DangerOfRunningAgroundData, vesselState);
				var warningManager = new WarningCollection();
				if (!warningManager.CloseProximityToLocation(newWarning))
				{
					warningManager.AddOrUpdate(newWarning);
					await hubContext.Clients.All.SendAsync("VesselInDangerOfRunningAground", newWarning);
				}
			}
		}


		/// <summary>
		/// A quick method that updates a VesselStates deviation data
		/// </summary>
		/// <param name="vesselState">The relevant VesselState</param>
		/// <param name="deviation">The deviatino containing data we want to insert into the VesselState</param>
		private static void UpdateVesselDeviation(VesselState vesselState, Deviation deviation)
		{
			vesselState.RelDistMax = deviation.rel_dist_max;
			vesselState.RelDistStd = deviation.rel_dist_std;
			vesselState.DistToP = deviation.disttoP;
		}

		/// <summary>
		/// Creates a thread that posts the VesselState to database and corresponding CurrentVesselState to clients
		/// </summary>
		/// <param name="hubContext">Interface connected to our hubs, allows the method to post to all clients</param>
		/// <param name="vesselState">The VesselState we are posting to our database</param>
		/// <param name="currentVesselState">The CurrentVesselState we are posting to clients</param>
		private void PostVesselStateToDBAndBroadcast(IHubContext<VesselHub> hubContext, VesselState vesselState, CurrentVesselState currentVesselState, IVesselStateRepository VesselStateRepository, bool isNewMessage = true)
		{
			#pragma warning disable CS4014 // Because this call is not awaited, execution of the current method continues before the call is completed
			Task.Run(() =>
			{
				currentVesselState.newMessage = isNewMessage;

				hubContext.Clients.All.SendAsync("broadcastUpdate", currentVesselState);
				lock (dbSyncRoot)
				{
					using (var repository = new VesselStateRepository())
					{
						if (vesselState.VesselStateId == 0)
					{
						repository.Add(vesselState);
					}
					else
					{
						repository.Update(vesselState);
					}
					repository.SaveChanges();
					}
				}
			});
		}

		/// <summary>
		/// Tries to update our CurrentVesselState dictionary
		/// </summary>
		/// <param name="vesselStateManager">The VesselStateManager containing our dictionary</param>
		/// <param name="NewCurrentVesselState">The new CurrentVesselState we received from stream</param>
		/// <param name="PreviousCurrentVesselState">The old CurrentVesselState we are checking against</param>
		/// <returns>true if the move is legal and update went successfully, false if not</returns>
		public bool TryUpdate(IVesselStateCollection vesselStateManager, CurrentVesselState NewCurrentVesselState, CurrentVesselState PreviousCurrentVesselState)
		{
			var update = false;
			if (PreviousCurrentVesselState == null)
			{
				NewCurrentVesselState.VesselTypeGroupId = GetVesselTypeGroupId(NewCurrentVesselState.MmsiNo);//DB query returns VesselTypeGroupeID
				NewCurrentVesselState.Description = GetDescription(NewCurrentVesselState.MmsiNo);
				if (NewCurrentVesselState.VesselTypeGroupId == 100232)
				{
					var typeId = GetVesselTypeId(NewCurrentVesselState.MmsiNo);//DB query returns VesselTypeID
					
					NewCurrentVesselState.VesselTypeId = typeId;
				}
				
				vesselStateManager.AddCurrentVesselState(NewCurrentVesselState);
				update = true;
			}
			else
			{
				update = VesselStateValidator.ValidNewState(NewCurrentVesselState, PreviousCurrentVesselState);
				if(update == true)
				{
					update = vesselStateManager.UpdateCurrentVesselState(NewCurrentVesselState);
				}
			}

			return update;
		}

		private string GetDescription(int MmsiNo)
		{
			return new VesselRepository().GetDescription(MmsiNo);
		}

		public int? GetVesselTypeGroupId(int MmsiNo)
		{
			var groupId = new VesselRepository().GetVesselTypeGroupId(MmsiNo);
			return groupId;
		}

		public int? GetVesselTypeId(int MmsiNo)
		{
			var typeId = new VesselRepository().GetVesselTypeId(MmsiNo);
			return typeId;
		}

		public class VesselStateValidator
		{

			/// <summary>
			/// Checks if the new position is legal. Illegal LatLong coordinates, too large or too small movements will be labeled as illegal.
			/// </summary>
			/// <param name="vesselState">The new vesselstate with new position</param>
			/// <returns>True if position is legal, false if not</returns>
			public static bool IsPositionValid(VesselState vesselState)
			{
				if (vesselState.Position.Y > 82 || vesselState.Position.Y < 55 || vesselState.Position.X > 40 || vesselState.Position.X < 0)
					return false;

				return true;
			}

			/// <summary>Checks if the distance between last position and new position is legal.</summary>
			/// <param name="newVesselState">The new vesselState with new position</param>
			/// <param name="previousVesselState"> The prevous vesselState we are comparing with</param>
			/// <returns>false if distance is too large or too small, true if not</returns>
			private static bool HasValidMovement(CurrentVesselState newVesselState, CurrentVesselState previousVesselState)
			{
				const double maxDistance = 1;
				const double minDistance = 0.0002;
				const int minSpeed = 0;

				// Speed should be above the minimum speed
				if (newVesselState.SpeedOverGround < minSpeed)
				{
					return false;
				}

				double distance = newVesselState.Position.Distance(previousVesselState.Position);
				if (distance >= maxDistance || distance < minDistance)
				{
					return false;
				}
				return true;
			}
			/// <summary> compares the time recieved of the new vessel state to the one in the currentVesselStates.</summary>
			/// <param name="currentVesselState">The new vessel state</param>
			/// <param name="previousVesselState">The vesselState that is currently in the dictionary</param>
			/// <returns>false if distance is too large or too small, true if not</returns>
			private static bool IsNew(CurrentVesselState currentVesselState, CurrentVesselState previousVesselState)
			{
				// Avoiding null pointer exceptions
				if (!currentVesselState.TimeReceived.HasValue)
					return false;

				if (!previousVesselState.TimeReceived.HasValue)
					return true;

				int result = DateTime.Compare(currentVesselState.TimeReceived.Value, previousVesselState.TimeReceived.Value);
				if (result > 0)
					return true;

				return false;
			}

			public static bool ValidNewState(CurrentVesselState currentVesselState, CurrentVesselState previousVesselState)
			{
				return HasValidMovement(currentVesselState, previousVesselState) && IsNew(currentVesselState, previousVesselState);
			}

		}

	}
}
